%
% 2013 Musa Maharramov
% Stanford Exploration Project
% SEP 150
%
% Demonstration of Robust Time-Lapse FWI
% using regularized simultaneous baseline and monitor
% model inversion and cross-updating
%
% Proof-of-concept "borehole" test problem (1D) 
%
% Forward-model 1D propagation in frequency domain
%
% Use Octave 3.6 or later, Matlab 7 or later
%
  close all; 
  clc;
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% PARAMETERS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% modeling and inversion grid size
  nx=32;

% flag to plot the 'parallel difference results'
% (that are typically the worst of all)
  PLOT_PARDIF=0;

% BASELINE AND MONITOR SOURCE PARAMETERS  
% Gaussian source deviation, coordinate, amplitude
  dev1=.03; sx1=.4;
  S1AMP=1.;
% source deviation, coordinate, amplitude
  dev2=.06; sx2=.6;
  S2AMP=1.;
%
% MODELING INTERVAL
  xleft=0; xright=1.;
%
% grid step
  dx=(xright-xleft)/(nx-1);
%
% BASELINE AND MONITOR RECEIVER ARRAY COORDINATES  
  rr1=linspace(xleft,xleft+dx*(nx-1),nx);
  rr2=linspace(xleft,xleft+dx*(nx-1),nx);
% perturb monitor receiver coordinates
% percentage coordinate perturbation
% REPRODUCIBILITY PRCV=0 PRCV=50
  txtRCV=getenv('SEP150PRCV');
  if isempty(txtRCV) 
    PRCV=50;
  else
    PRCV=str2num(txtRCV);
  end
% set the random number generator to a fixed state
%  -- solely for plot reproducibility
% Octave only
%  state=[200]; rand('state',state);
% on Matlab comment out the above and use 
  rng(200);
%
  rr2=rr2+[0 PRCV/100*rand(1,nx-2)*dx 0];

%
% PRODUCTION-INDUCED PERTURBATION (in monitor)
% perturbation depth, width and magnitude
  PERTD=.25; PERTW=.4; PERTM=.05;

% MODELING/MULTI-SCALE INVERSION FREQUENCIES
% pick frequencies to achieve the required 
% model resolution, choose low (~1Hz) 
% upper bound to simulate numerical repeatability
% issues - i.e., spurious and different inversion
% artifacts in the baseline and monitor inversions
  fmin=.2;df=.2;fmax=1.;
%
% Percentage of incoherent noise to add above 
% PNYQ% of Nyquist
% NOTE: if increasing 'fmax', reduce 'PNOISE'
%       or increase 'PNYQ'
% REPRODUCIBILITY: PNOISE=0 PNOISE=5.
  txtPNOISE=getenv('SEP150PNOISE');
  if isempty(txtPNOISE)
    PNOISE=5.;
  else
    PNOISE=str2num(txtPNOISE);
  end
  PNYQ=80.;
%
% END OF PARAMETER SETTING
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% INITIALIZATION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
%
% plot name prefix
  PREFIX=...
 ['N' num2str(nx)  ...
  'D' num2str(dx)  ...
  'F0' num2str(fmin)  ...
  'dF' num2str(df)  ...
  'FM' num2str(fmax)  ...
  'PNS' num2str(PNOISE)  ...
  'PNQ' num2str(PNYQ)  ...
  'PRCV' num2str(PRCV)  ...
  'SX1' num2str(sx1)  ...
  'DEV1' num2str(dev1)  ...
  'S1A' num2str(S1AMP)  ...
  'SX2' num2str(sx2)  ...
  'DEV2' num2str(dev2)  ...
  'S2A' num2str(S2AMP)  ...
  'PPD' num2str(PLOT_PARDIF) ];


% set up modeling and inversion grid 
  xx=linspace(xleft,xleft+dx*(nx-1),nx);
  dx2=dx^2;
%
% baseline source
  s1=S1AMP*exp(-((xx-sx1).^2)/dev1^2)';
% true baseline model
  v1=[.4*ones(nx/2+nx/4,1); .43*ones(nx/4,1)]+ ...
     linspace(-.05,.05,nx)';
% number of baseline receivers
  nr1=length(rr1);
% monitor source 
  s2=S2AMP*exp(-((xx-sx2).^2)/dev2^2)';
% true monitor model
% monitor perturbation depth, width on grid
  pertdi=floor(PERTD/dx)+1; pertwi=floor(PERTW/dx); 
  v2=v1+smooth_pert(nx,pertdi,pertwi,PERTM);
% number of monitor receivers
  nr2=length(rr2);
% modeling/inversion frequencies
  freq=2*pi*[fmin:df:fmax]';
  Nfreq=length(freq);
%
% create measurement operators 
% for baseline and monitor
  M1=receiver_measurement_op(xx,rr1);
  M2=receiver_measurement_op(xx,rr2);
% cross-survey interpolation operator
  MI=receiver_measurement_op(rr2,rr1);
%
  DATA1=zeros(nr1, Nfreq); DATA2=zeros(nr2, Nfreq);
%

%
% FORWARD-MODELING
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
%
%
  for ifreq=1:Nfreq
    DATA1(:,ifreq)=M1*forward_model(nx,dx2,v1.^2,freq(ifreq)^2,s1);
    DATA2(:,ifreq)=M2*forward_model(nx,dx2,v2.^2,freq(ifreq)^2,s2);
  end
%
% ADD PNOISE% COLORED NOISE ABOVE
% PNYQ% OF NYQUIST WN
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
  for ifreq=1:Nfreq
% generated colored noise above method resolution
    amp=norm(DATA1(:,ifreq),inf);
    noise=amp*rand(nr1,1);
    fnoise=fft(noise);
%   hi-pass
    NLOW=floor((1-PNYQ/100.)*nr1/2);
    fnoise(1:NLOW)=0; fnoise(nr1-NLOW+2:nr1)=0;
    noise=ifft(fnoise);
    DATA1(:,ifreq)=DATA1(:,ifreq)+PNOISE*noise/100.;
    amp=norm(DATA2(:,ifreq),inf);
    noise=amp*rand(nr2,1);
    fnoise=fft(noise);
%   hi-pass
    NLOW=floor((1-PNYQ/100.)*nr2/2);
    fnoise(1:NLOW)=0; fnoise(nr2-NLOW+2:nr2)=0;
    noise=ifft(fnoise);
    DATA2(:,ifreq)=DATA2(:,ifreq)+PNOISE*noise/100.;
  end
%
% INVERSION PARAMETERS AND RESULTS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
% results of multi-scale inversion
% for the baseline and monitor
  v1o=zeros(nx,Nfreq); v2o=zeros(nx,Nfreq);
%
% keep the number of iterations small to demonstrate the
% advantages of simultaneous inversion
  NITER=100;
%
% SEPARATE BASELINE/MONITOR INVERSION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
%
% Multi-scale baseline model inversion
  v1_out=.4^2*ones(length(v1),1);
  for ifreq=1:Nfreq
    PAR.nvar=length(v1);
%   name of the function that computes objective function 
%   value and gradient with respect to the velocity
    PAR.fgname='ofg';
%   note we pass additional modeling parameters
%   via the 'PAR' structure
    PAR.dx2=dx2;
    PAR.M=M1;
    PAR.freq2=freq(ifreq)^2;
    PAR.s=s1;
    PAR.data=DATA1(:,ifreq);
% initial approximation
    PAR.x0=v1_out;
    PAR.version='C';
    PAR.maxit=NITER;
    PAR.normtol=1.e-20;
%
    disp(['Inverting the baseline for freq=' num2str(freq(ifreq))]);
    [v1_out, f1_out]=ncg(PAR);
    v1o(:,ifreq)=v1_out;
    clear PAR PAR;
  end
%
% same for the monitor
  v2_out=.4^2*ones(length(v2),1);
  for ifreq=1:Nfreq
    PAR.nvar=length(v2);
    PAR.fgname='ofg';
    PAR.dx2=dx2;
    PAR.M=M2;
    PAR.freq2=freq(ifreq)^2;
    PAR.s=s2;
    PAR.data=DATA2(:,ifreq);
% initial approximation
    PAR.x0=v2_out;
    PAR.maxit=NITER;
    PAR.normtol=1.e-20;
%
    disp(['Inverting the baseline for freq=' num2str(freq(ifreq))]);
    [v2_out, f2_out]=ncg(PAR);
    v2o(:,ifreq)=v2_out;
    clear PAR PAR;
  end
%


% SEQUENTIAL DIFFERENCE 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
  v22_out=v1_out;
  for ifreq=1:Nfreq
    PAR.nvar=length(v2);
    PAR.fgname='ofg';
    PAR.dx2=dx2;
    PAR.M=M2;
    PAR.freq2=freq(ifreq)^2;
    PAR.s=s2;
    PAR.data=DATA2(:,ifreq);
% initial approximation
    PAR.x0=v22_out;
    PAR.maxit=NITER;
    PAR.normtol=1.e-20;
%
    disp(['SEQ Inverting the baseline for freq=' num2str(freq(ifreq))]);
    [v22_out, f2_out]=ncg(PAR);
    clear PAR PAR;
  end
%

% X-UPDATING 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
% new baseline inversion starting from the sequentially 
% inverted monitor
  v13_out=v22_out;
  for ifreq=1:Nfreq
    PAR.nvar=length(v1);
%   name of the function that computes objective function 
%   value and gradient with respect to the velocity
    PAR.fgname='ofg';
%   note we pass additional modeling parameters
%   via the 'PAR' structure
    PAR.dx2=dx2;
    PAR.M=M1;
    PAR.freq2=freq(ifreq)^2;
    PAR.s=s1;
    PAR.data=DATA1(:,ifreq);
% initial approximation
    PAR.x0=v13_out;
    PAR.maxit=NITER;
    PAR.normtol=1.e-20;
%
    disp(['Inverting the baseline for freq=' num2str(freq(ifreq))]);
    [v13_out, f1_out]=ncg(PAR);
    clear PAR PAR;
  end
%
% monitor starting from the new baseline
  v23_out=v13_out;
  for ifreq=1:Nfreq
    PAR.nvar=length(v2);
    PAR.fgname='ofg';
    PAR.dx2=dx2;
    PAR.M=M2;
    PAR.freq2=freq(ifreq)^2;
    PAR.s=s2;
    PAR.data=DATA2(:,ifreq);
% initial approximation
    PAR.x0=v23_out;
    PAR.maxit=NITER;
    PAR.normtol=1.e-20;
%
    disp(['Inverting the baseline for freq=' num2str(freq(ifreq))]);
    [v23_out, f2_out]=ncg(PAR);
    clear PAR PAR;
  end
%
% BETTER INITIAL MODELS 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
% new baseline inversion starting from the sequentially 
% inverted monitor
  v14_out=v1_out;
  for ifreq=1:Nfreq
    PAR.nvar=length(v1);
%   name of the function that computes objective function 
%   value and gradient with respect to the velocity
    PAR.fgname='ofg';
%   note we pass additional modeling parameters
%   via the 'PAR' structure
    PAR.dx2=dx2;
    PAR.M=M1;
    PAR.freq2=freq(ifreq)^2;
    PAR.s=s1;
    PAR.data=DATA1(:,ifreq);
% initial approximation
    PAR.x0=v14_out;
    PAR.maxit=NITER;
    PAR.normtol=1.e-20;
%
    disp(['Inverting the baseline for freq=' num2str(freq(ifreq))]);
    [v14_out, f1_out]=ncg(PAR);
    clear PAR PAR;
  end
%
% monitor starting from the new baseline
  v24_out=v22_out;
  for ifreq=1:Nfreq
    PAR.nvar=length(v2);
    PAR.fgname='ofg';
    PAR.dx2=dx2;
    PAR.M=M2;
    PAR.freq2=freq(ifreq)^2;
    PAR.s=s2;
    PAR.data=DATA2(:,ifreq);
% initial approximation
    PAR.x0=v24_out;
    PAR.maxit=NITER;
    PAR.normtol=1.e-20;
%
    disp(['Inverting the baseline for freq=' num2str(freq(ifreq))]);
    [v24_out, f2_out]=ncg(PAR);
    clear PAR PAR;
  end
%

  L3=sparse(-diag(ones(nx-1,1),1)-diag(ones(nx-1,1),-1)+2*diag(ones(nx,1)));
  L3(1,:)=0; L3(nx:0)=0;
%
  L1=L3; L2=L3;
  
% SIMULTANEOUS BASELINE & MONITOR INVERSION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
%
%
% regularization parameters
  ALPHAREG=linspace(0.,.0,Nfreq);
  BETAREG=linspace(0.,.0,Nfreq);
% use only difference terms
%  GAMMAREG=linspace(1,1,Nfreq);
% use stronger smoothing with noise
  GAMMAREG=linspace(2,.2,Nfreq);

  MM=[M1 zeros(nr1,nx);
      zeros(nr2,nx) M2];
%  vv_out=.65^2*ones(length(v1)+length(v2),1);
  vv_out=[v14_out; v24_out];
  for ifreq=1:Nfreq
    PAR.nvar=length(v1)+length(v2);
%   name of the function that computes objective function 
%   value and gradient with respect to the velocity
    PAR.fgname='ofgSimu';
%   note we pass additional modeling parameters
%   via the 'PAR' structure
    PAR.dx2=dx2;
    PAR.M=MM;
    PAR.freq2=freq(ifreq)^2;
    PAR.s=[s1; s2];
    PAR.nr1=nr1;
    PAR.nr2=nr2;
    PAR.data=[DATA1(:,ifreq); 
               DATA2(:,ifreq)];
    PAR.R=[ALPHAREG(ifreq)*[L1 zeros(nx,nx)];
             BETAREG(ifreq)*[zeros(nx,nx) L2];
             GAMMAREG(ifreq)*[L3 (-1)*L3]];
%   prior geophysical knowledge
    PAR.R(2*nx+pertdi+1:2*nx+pertdi+pertwi-1,:)=0;
    PAR.prior=[zeros(nx,1); zeros(nx,1); zeros(nx,1)];
% initial approximation
    PAR.x0=vv_out;
    PAR.maxit=NITER;
    PAR.normtol=1.e-20;
%
    disp(['Simultaneously inverting baseline and monitor for freq=' num2str(freq(ifreq))]);
    [vv_out, f1_out]=ncg(PAR);
    clear PAR;
  end


% OUTPUT
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
%

figure
  plot(xx,v1,'-+k','LineWidth',2);
  hold on;
  plot(xx,v14_out.^.5,'--c','LineWidth',2);
  hold on;
  plot(xx,v13_out.^.5,'-g','LineWidth',2);
  hold on;
  plot(xx,vv_out(1:nx).^.5,'-r','LineWidth',2);
  hold off;
  title('Inverted Baseline','FontSize',12);
  legend('true','sequential','X-updated','simultaneous','Location','SouthEast');
  xlabel('x','fontsize',12,'fontweight','b','color','k');
  ylabel('v','fontsize',12,'fontweight','b','color','k');
  set(gcf, 'PaperUnits', 'inches');
  set(gcf, 'PaperSize', [6 5]);
  set(gcf, 'PaperPositionMode', 'auto');  
  print(gcf,'-dpdf',[PREFIX 'baseline.pdf']);

figure
  plot(xx,v2,'-+k','LineWidth',2);
  hold on;
  plot(xx,v24_out.^.5,'--c','LineWidth',2);
  hold on;
  plot(xx,v23_out.^.5,'-g','LineWidth',2);
  hold on;
  plot(xx,vv_out(nx+1:2*nx).^.5,'-r','LineWidth',2);
  hold off;
  title('Inverted Monitor','FontSize',12);
  legend('true','sequential','X-updated','simultaneous','Location','SouthEast');
  xlabel('x','fontsize',12,'fontweight','b','color','k');
  ylabel('v','fontsize',12,'fontweight','b','color','k');
  set(gcf, 'PaperUnits', 'inches');
  set(gcf, 'PaperSize', [6 5]);
  set(gcf, 'PaperPositionMode', 'auto');  
  print(gcf,'-dpdf',[PREFIX 'monitor.pdf']);

figure
  plot(xx,v2-v1,'-+k','LineWidth',2);
  hold on;
  plot(xx,v24_out.^.5-v14_out.^.5,'--c','LineWidth',2);
  hold on;
  plot(xx,v23_out.^.5-v13_out.^.5,'-g','LineWidth',2);
  hold on;
  plot(xx,vv_out(nx+1:2*nx).^.5-vv_out(1:nx).^.5,'-r','LineWidth',2);
  hold off;
  title('Inverted Model Difference','FontSize',12);
  legend('true',...
  'sequential','X-updated',...
  'simultaneous','Location','NorthWest');
  xlabel('x','fontsize',12,'fontweight','b','color','k');
  ylabel('v','fontsize',12,'fontweight','b','color','k');
  set(gcf, 'PaperUnits', 'inches');
  set(gcf, 'PaperSize', [6 5]);
  set(gcf, 'PaperPositionMode', 'auto');  
  print(gcf,'-dpdf',[PREFIX 'difference.pdf']);


% smoothing parameter
SMOOTHW=pertwi/2;
if mod(SMOOTHW,2)==0
  SMOOTHW=SMOOTHW+1;
end

figure
  plot(xx,v2-v1,'-+k','LineWidth',2);
  hold on;
  plot(xx,sgolayfilt(v24_out.^.5-v14_out.^.5,3,SMOOTHW),'--c','LineWidth',2);  
  hold on;
  plot(xx,sgolayfilt(v23_out.^.5-v13_out.^.5,3,SMOOTHW),'-g','LineWidth',2);  
  hold on;
  plot(xx,sgolayfilt(vv_out(nx+1:2*nx).^.5-vv_out(1:nx).^.5,3,SMOOTHW),'-r','LineWidth',2);
  hold off;
  title('Smoothed Inverted Model Difference','FontSize',12);
  legend('true',...
  'sequential','X-updated',...
  'simultaneous','Location','NorthWest');
  xlabel('x','fontsize',12,'fontweight','b','color','k');
  ylabel('v','fontsize',12,'fontweight','b','color','k');
  set(gcf, 'PaperUnits', 'inches');
  set(gcf, 'PaperSize', [6 5]);
  set(gcf, 'PaperPositionMode', 'auto');  
  print(gcf,'-dpdf',[PREFIX 'smootheddifference.pdf']);

figure
  plot(xx,s1,'-r','LineWidth',2);
  hold on;
  plot(xx,s2,'-g','LineWidth',2);
  hold off;
  title('Baseline and Monitor Survey Sources','FontSize',12);
  legend('baseline','monitor','Location','NorthWest');
  xlabel('x','fontsize',12,'fontweight','b','color','k');
  ylabel('amplitude','fontsize',12,'fontweight','b','color','k');
  set(gcf, 'PaperUnits', 'inches');
  set(gcf, 'PaperSize', [6 5]);
  set(gcf, 'PaperPositionMode', 'auto');  
  print(gcf,'-dpdf',[PREFIX 'sources.pdf']);

figure
  plot(xx,v1,'-k','LineWidth',2);
  title('True Baseline Model','FontSize',12);
  xlabel('x','fontsize',12,'fontweight','b','color','k');
  ylabel('v','fontsize',12,'fontweight','b','color','k');
  set(gcf, 'PaperUnits', 'inches');
  set(gcf, 'PaperSize', [6 5]);
  set(gcf, 'PaperPositionMode', 'auto');  
  print(gcf,'-dpdf',[PREFIX 'truebaseline.pdf']);

figure
  plot(xx,v2,'-k','LineWidth',2);
  title('True Monitor Model','FontSize',12);
  xlabel('x','fontsize',12,'fontweight','b','color','k');
  ylabel('v','fontsize',12,'fontweight','b','color','k');
  set(gcf, 'PaperUnits', 'inches');
  set(gcf, 'PaperSize', [6 5]);
  set(gcf, 'PaperPositionMode', 'auto');  
  print(gcf,'-dpdf',[PREFIX 'truemonitor.pdf']);
%
quit

